import pytest

from rumydata import table, Layout, field
from rumydata.rules import row as rr, header as hr


def test_row_good(basic):
    assert not table.Layout(basic).check_row(['1', '2', '2020-01-01', 'X'])


def test_row_choice(basic):
    lay = Layout({'c1': field.Choice(['x'], nullable=True)})
    assert not lay.check_row(['x'])
    assert not lay.check_row([''])


@pytest.mark.parametrize('value,err', [
    ([1, 2, 3, 4, 5], rr.RowLengthLTE),
    ([1, 2, 3], rr.RowLengthGTE)
])
def test_row_bad(basic, value, err):
    assert table.Layout(basic)._has_error(value, err.rule_exception(), rule_type=rr.Rule)


def test_header_good(basic):
    assert not table.Layout(basic).check_header(['col1', 'col2', 'col3', 'col4'])


@pytest.mark.parametrize('value,err', [
    (['col1', 'col2'], hr.NoMissing),
    (['col1', 'col2', 'col2'], hr.NoDuplicate),
    (['col1', 'col2', 'col5'], hr.NoExtra)
])
def test_header_bad(basic, value, err):
    assert table.Layout(basic)._has_error(value, err.rule_exception(), rule_type=hr.Rule)


def test_header_skip(basic):
    assert not table.Layout(basic, skip_header=True). \
        check_header(['col1', 'col2', 'col4'])


@pytest.mark.parametrize('row_length,row,expected', [
    (1, ['1', '2'], False),
    (1, ['1'], True)
])
def test_row_length_lte(row_length, row, expected):
    r = rr.RowLengthLTE(row_length)
    assert r._evaluator()(*r._prepare(row)) is expected


@pytest.mark.parametrize('row_length,row,expected', [
    (3, ['1', '2'], False),
    (2, ['1', '2'], True),
    (1, ['1', '2'], True)
])
def test_row_length_gte(row_length, row, expected):
    r = rr.RowLengthGTE(row_length)
    assert r._evaluator()(*r._prepare(row)) is expected



"""
Exception module

This module contains specialized Exception classes which allow for aggregation
of all errors identified in a subject. These exceptions provide the structure to
collect multiple exceptions at varying levels within a validation process, and
display them in a meaningful way.
"""


def debug():
    """
    Exception debug mode control

    A switch to control exception message sanitation. This is a static
    function which will always return False. In order to enable debug mode, you
    must monkey patch this function using the mock library. This is made to
    intentionally be inconvenient in order to prevent accidental disclosure of
    data in exception logs through overuse of the debug mode.

    This method is used at runtime by the base _check method. In order to enable
    debugging, you must include a patch statement before calling a check. It is
    recommended that you do this with a context handler to ensure appropriate
    behavior of the patch.

    example::

        from unittest.mock import patch
        from rumydata import field

        with patch('rumydata.exception.debug', return_value=True):
            field.Integer(1).check_cell(None)
    """
    return False


class UrNotMyDataError(Exception):
    """
    Base exception class for rumydata package

    This exception provides the structure for reporting nested exceptions in
    text format, in a way that the nested structure can be interpreted.

    The constructor allows a list of errors to be associated with this
    exception, in addition to providing a way to override the default
    message printed when this exception is raised.

    :param msg: a custom message to override the default with this exception
        is raised
    :param errors: a list of UrNotMyDataError exceptions which are nested
        within this object.
    """

    _message: str = None

    def __init__(self, msg: str = None, errors: list = None):
        super().__init__(msg)
        self._message = msg or self._message
        self._errors = errors or []

    def __str__(self) -> str:
        """
        Exception string override

        This method overrides the default string and instead replaces it with
        the `md` method in this object. This allows simple string printing that
        can represent nested exceptions that exist in this object, and the
        exceptions that exist in those objects, and so on.
        """

        return '\n' + self._md()

    def _md(self, depth=0) -> str:
        """
        Nested exception Markdown digest

        This method returns the complete tree of nested exceptions that exist in
        the errors property of this class in Markdown format, with indentation
        provided to indicate the relationship of nested exceptions.

        :param depth: an integer which indicates the level of indentation that
            an exception needs to represent its relationship in the nested
            structure. Do not call directly.
        :return: a string of nested exceptions in Markdown format, with
            indentation providing visual indicator of nested structure.
        """

        txt = f'{"  " * depth} - {self.__class__.__name__[:-5]}: {self._message}'
        if self._errors:
            txt = '\n'.join([txt] + [x for x in self._flatten_md(self._errors, depth)])
        return txt

    @classmethod
    def _flatten_md(cls, errors, depth=0):
        """
        Nested exception Markdown generator

        Recurse through nested exceptions and continue to yield exceptions until
        exhausted.
        """

        depth += 1
        for el in errors:
            yield el._md(depth)


class FileError(UrNotMyDataError):
    """
    File Error exception

    :param file: name of the file reported with the error message.
    :param msg: a custom message to use when reporting errors in the file.
    :param errors: a list of errors contained in the file.
    """

    def __init__(self, file, msg=None, errors: list = None):
        message = file
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class ColumnError(UrNotMyDataError):
    """
    Column Error Exception

    :param msg: a custom message to use when reporting errors in the column.
    :param errors: a list of errors contained in the column.
    """

    def __init__(self, index: int, msg=None, errors: list = None, **kwargs):
        message = ''
        offset = 0 if kwargs.get("zero_index") else 1

        message += f'{str(index + offset)}'
        if kwargs.get("name"):
            message += f' ({kwargs.get("name")})'
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class RowError(UrNotMyDataError):
    """
    Row Error exception

    :param index: row number (using zero-indexing). Reported with the
        exception message to allow identification of the row.
    :param msg: a custom message to use when reporting errors in this row.
    :param errors: a list of errors contained in this row
    :param zero_index: specify whether to use zero-indexing when reporting
        row number. If True, the index is reported as provided. If False,
        the index is increased by one.
    """

    def __init__(self, index: int, msg=None, errors: list = None, **kwargs):
        message = f'{str(index + (0 if kwargs.get("zero_index") else 1))}'
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class CellError(UrNotMyDataError):
    """
    Cell Error exception

    :param index: column number (using zero-indexing). Reported with the
        exception message to allow identification of the column.
    :param msg: a custom message to use when reporting errors in this cell.
    :param errors: a list of errors contained in this cell.
    :param zero_index: specify whether to use zero-indexing when reporting
        column number. If True, the index is reported as provided. If False,
        the index increased by one.
    """

    def __init__(self, index: int, msg=None, errors: list = None, **kwargs):
        message = ''
        offset = 0 if kwargs.get("zero_index") else 1
        if kwargs.get('rix'):
            message = str(kwargs.get('rix') + offset) + ','

        message += f'{str(index + offset)}'
        if kwargs.get("name"):
            message += f' ({kwargs.get("name")})'
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class PreProcessingError(UrNotMyDataError):
    """
    Data Preprocessing Exception

    Thrown specifically when the data pre-processing step for a subject fails.
    This is treated differently than the exceptions thrown by all fields, since
    pre-processing exceptions prevent any further rules from being checked.
    """



from datetime import date
from pathlib import Path

import pandas as pd
import pytest

import rumydata.rules.cell
import rumydata.table
from rumydata import field
from rumydata.table import ParquetFile
from tests.utils import mock_no_module


@pytest.fixture()
def basic_good(tmpdir):
    p = Path(tmpdir, 'good.parquet')
    df = pd.DataFrame({
        'col1': ['A'],
        'col2': [1],
        'col3': [date(2020, 1, 1)],
        'col4': ['Z']
    })
    df.to_parquet(p)
    yield p.as_posix()


@pytest.fixture()
def basic_bad(tmpdir):
    p = Path(tmpdir, 'good.parquet')
    df = pd.DataFrame({
        'col1': ['A'],
        'col2': [1],
        'col3': [date(2020, 1, 1)],
        'col4': ['z']
    })
    df.to_parquet(p)
    yield p.as_posix()


def test_file_good(basic_good, basic):
    assert not ParquetFile(rumydata.table.Layout(basic)).check(basic_good)


def test_file_bad(basic_bad, basic):
    assert ParquetFile(rumydata.table.Layout(basic))._has_error(basic_bad, rumydata.rules.cell.Choice.rule_exception())


def test_no_pandas(mocker):
    mocker.patch('builtins.__import__', wraps=__import__, side_effect=mock_no_module('pandas'))
    with pytest.raises(ModuleNotFoundError):
        ParquetFile(rumydata.table.Layout({'x': field.Integer(1)}))


def test_no_pyarrow(mocker):
    mocker.patch('builtins.__import__', wraps=__import__, side_effect=mock_no_module('pyarrow'))
    with pytest.raises(ModuleNotFoundError):
        ParquetFile(rumydata.table.Layout({'x': field.Integer(1)}))



from datetime import datetime as dt

import pytest

from rumydata import field, rules


def recurse_subclasses(class_to_recurse):
    def generator(x):
        for y in x.__subclasses__():
            for z in generator(y):
                yield z
        yield x

    return list(generator(class_to_recurse))


@pytest.mark.parametrize('fo,rule', [
    (field.Text(1), rules.cell.MaxChar),
    (field.Text(1, 2), rules.cell.MinChar),
    (field.Text(1, rules=[rules.cell.AsciiChar()]), rules.cell.AsciiChar)
])
def test_has_rule(fo, rule):
    assert fo._has_rule_type(rule)


@pytest.mark.parametrize('fo,columns', [
    (field.Field(rules=[rules.cell.GreaterThanColumn('x')]), set('x')),
    (field.Field(
        rules=[rules.cell.GreaterThanColumn('x'), rules.cell.GreaterThanColumn('y')]
    ), {'x', 'y'}),
    (field.Field(rules=[]), set())
])
def test_comparison_columns(fo, columns: set):
    assert fo._comparison_columns() == columns


@pytest.mark.parametrize('fo', recurse_subclasses(field.Field))
def test_digest(fo):
    assert isinstance(fo(*fo._default_args)._digest(), list)


@pytest.mark.parametrize('value,kwargs', [
    ('x', dict(max_length=1)),
    ('x', dict(max_length=2)),
    ('', dict(max_length=1, nullable=True)),
    ('', dict(max_length=1, min_length=1, nullable=True))
])
def test_text_good(value, kwargs):
    assert not field.Text(**kwargs).check_cell(value)


@pytest.mark.parametrize('value,kwargs,rule', [
    ('', dict(max_length=1), rules.cell.NotNull),
    ('xxx', dict(max_length=2), rules.cell.MaxChar),
    ('x', dict(max_length=80, min_length=2), rules.cell.MinChar),
])
def test_text_bad(value, kwargs, rule):
    assert field.Text(**kwargs)._has_error(value, rule.rule_exception())


@pytest.mark.parametrize('value,kwargs', [
    ('2020-01-01', {}),
    ('', dict(nullable=True)),
    ('2020-01-01', dict(max_date='2020-01-01')),
    ('2020-01-01', dict(max_date='2020-01-02')),
    ('2020-01-01', dict(min_date='2020-01-01', max_date='2020-01-02')),
    ('2020-01-01 00:00:00', dict(truncate_time=True))
])
def test_date_good(value, kwargs):
    assert not field.Date(**kwargs).check_cell(value)


@pytest.mark.parametrize('value,rule,kwargs', [
    ('', rules.cell.NotNull, {}),
    ('20200101', rules.cell.CanBeDateIso, {}),
    ('9999-99-99', rules.cell.CanBeDateIso, {}),
    ('2020-01-01', rules.cell.DateGTE, dict(min_date='2020-01-02')),
    ('2020-01-02', rules.cell.DateLTE, dict(max_date='2020-01-01')),
    ('2020-01-01', rules.cell.DateGTE, dict(min_date='2020-01-02', max_date='2020-01-03')),
    ('2020-01-05', rules.cell.DateLTE, dict(min_date='2020-01-02', max_date='2020-01-03')),
    ('2020-01-01 00:00:00', rules.cell.CanBeDateIso, dict(truncate_time=True)),
    ('2020-01-01 00:00:01', rules.cell.CanBeDateIso, dict(truncate_time=False))
])
def test_date_bad(value, rule, kwargs):
    assert field.Date(**kwargs)._has_error(value, rule.rule_exception())


@pytest.mark.parametrize('value,sig_dig,kwargs', [
    ('123.45', 5, {}),
    ('123.00', 5, {}),
    ('123.0', 5, {}),
    ('123', 5, {}),
    ('', 1, dict(nullable=True)),
    ('-0.01', 3, dict(rules=[rules.cell.NumericLT(0)])),
    ('0', 3, dict(rules=[rules.cell.NumericLTE(0)])),
    ('0.00', 3, dict(rules=[rules.cell.NumericET(0)])),
    ('0', 3, dict(rules=[rules.cell.NumericGTE(0)])),
    ('0.01', 3, dict(rules=[rules.cell.NumericGT(0)])),
])
def test_currency_good(value, sig_dig, kwargs):
    assert not field.Currency(sig_dig, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,sig_dig,rules_list,err', [
    ('0.00', 5, [rules.cell.NumericLT(0)], rules.cell.NumericLT),
    ('0.01', 5, [rules.cell.NumericLTE(0)], rules.cell.NumericLTE),
    ('0.01', 5, [rules.cell.NumericET(0)], rules.cell.NumericET),
    ('-0.01', 5, [rules.cell.NumericGTE(0)], rules.cell.NumericGTE),
    ('-0.01', 5, [rules.cell.NumericGT(0)], rules.cell.NumericGT),
    ('', 5, [], rules.cell.NotNull),
    ('123.45', 4, [], rules.cell.MaxDigit),
    ('123.', 4, [], rules.cell.NumericDecimals),
    ('123.456', 4, [], rules.cell.NumericDecimals)
])
def test_currency_bad(value, sig_dig, rules_list, err):
    assert field.Currency(sig_dig, rules=rules_list)._has_error(value, err.rule_exception())


@pytest.mark.parametrize('value,max_length, kwargs', [
    ('1', 3, {}),
    ('12', 3, {}),
    ('123', 3, {}),
    ('12', 2, dict(min_length=2)),
    ('123', 3, dict(min_length=2))
])
def test_digit_good(value, max_length, kwargs):
    assert not field.Digit(max_length, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,max_length,err,kwargs', [
    ('-123', 3, rules.cell.OnlyNumbers, {}),
    ('-123', 3, rules.cell.MaxChar, {}),
    ('1', 2, rules.cell.MinChar, dict(min_length=2)),
    ('123456', 3, rules.cell.MaxChar, dict(min_length=2))
])
def test_digit_bad(value, max_length, err, kwargs):
    assert field.Digit(max_length, **kwargs)._has_error(value, err.rule_exception())


@pytest.mark.parametrize('value,max_length,kwargs', [
    ('-1', 1, {}),
    ('0', 1, {}),
    ('1', 1, {}),
    ('1', 2, {}),
    ('11', 2, dict(min_length=2)),
    ('', 1, dict(nullable=True)),
    ('-1', 1, dict(rules=[rules.cell.NumericLT(0)])),
    ('-1', 1, dict(rules=[rules.cell.NumericLTE(0)])),
    ('0', 1, dict(rules=[rules.cell.NumericLTE(0)])),
    ('0', 1, dict(rules=[rules.cell.NumericET(0)])),
    ('0', 1, dict(rules=[rules.cell.NumericGTE(0)])),
    ('1', 1, dict(rules=[rules.cell.NumericGTE(0)])),
    ('1', 1, dict(rules=[rules.cell.NumericGT(0)]))
])
def test_integer_good(value, max_length, kwargs):
    assert not field.Integer(max_length, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,max_length,kwargs,err', [
    ('0', 1, dict(rules=[rules.cell.NumericLT(0)]), rules.cell.NumericLT),
    ('1', 1, dict(rules=[rules.cell.NumericLTE(0)]), rules.cell.NumericLTE),
    ('1', 1, dict(rules=[rules.cell.NumericET(0)]), rules.cell.NumericET),
    ('-1', 1, dict(rules=[rules.cell.NumericGTE(0)]), rules.cell.NumericGTE),
    ('0', 1, dict(rules=[rules.cell.NumericGT(0)]), rules.cell.NumericGT),
    ('', 1, {}, rules.cell.NotNull),
    ('1', 2, dict(min_length=2), rules.cell.MinDigit),
    ('111', 2, {}, rules.cell.MaxDigit),
    ('00', 2, {}, rules.cell.NoLeadingZero),
    ('01', 2, {}, rules.cell.NoLeadingZero)
])
def test_integer_bad(value, max_length, kwargs, err):
    assert field.Integer(max_length, **kwargs)._has_error(value, err.rule_exception())


@pytest.mark.parametrize('value,choices,kwargs', [
    ('x', ['x'], {}),
    ('x', ['x', 'y'], {}),
    ('y', ['x', 'y'], {}),
    ('', ['x'], dict(nullable=True)),
    ('X', ['x'], dict(case_insensitive=True)),
    ('x', ['X'], dict(case_insensitive=True))
])
def test_choice_good(value, choices, kwargs):
    assert not field.Choice(choices, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,choices,kwargs,err', [
    ('', ['x'], {}, rules.cell.NotNull),
    ('x', ['z'], {}, rules.cell.Choice)
])
def test_choice_bad(value, choices, kwargs, err):
    assert field.Choice(choices, **kwargs)._has_error(value, err.rule_exception())


@pytest.mark.parametrize('value,func,assertion', [
    ('2', lambda x: int(x) % 2 == 0, "must be an even number"),
    ('1', lambda x: int(x) % 2 == 1, "must be an odd number"),
    ('2020-01-01', lambda x: dt.fromisoformat(x), "must be an isodate"),
])
def test_static_rules_good(value, func, assertion):
    x = field.Field(rules=[rules.cell.make_static_cell_rule(func, assertion)])
    assert not x.check_cell(value)


@pytest.mark.parametrize('value,func,assertion', [
    ('2', lambda x: int(x) % 2 == 1, "must be an even number"),
    ('1', lambda x: int(x) % 2 == 0, "must be an odd number"),
    ('2020-00-01', lambda x: dt.fromisoformat(x), "must be an isodate"),
    ('', lambda x: 1 / 0, "custom exception")
])
def test_static_rules_bad(value, func, assertion):
    r = rules.cell.make_static_cell_rule(func, assertion)
    x = field.Field(rules=[r])
    assert x._has_error(value, r.rule_exception())


@pytest.mark.parametrize('cell', [
    '',
    '1',
    '8k;abc;abc'
])
def test_ignore_cell(cell):
    assert not field.Ignore().check_cell(cell)


def test_column_compare_rule_good():
    x = field.Field(rules=[rules.cell.GreaterThanColumn('x')])
    assert not x.check_cell(('1', {'x': '0'}))


def test_column_compare_rule_bad():
    x = field.Field(rules=[rules.cell.GreaterThanColumn('x')])
    assert x._has_error('1', compare={'x': '1'}, error=rules.cell.GreaterThanColumn.rule_exception())


def test_column_unique_good():
    x = field.Field(rules=[rules.column.Unique()])
    assert not x.check_column(['1', '2', '3'])


def test_column_unique_bad():
    x = field.Field(rules=[rules.column.Unique()])
    assert x._has_error(['2', '2'], rules.column.Unique.rule_exception(), rule_type=field.cr.Rule)


def test_empty_field():
    empty = field.Empty()
    assert not empty.check_cell('')
    with pytest.raises(AssertionError):
        assert empty.check_cell('1')



from pathlib import Path
from uuid import uuid4

import pytest
from openpyxl import Workbook

from rumydata.field import Integer, Field, Text
from rumydata.rules.column import Unique
from rumydata.table import Layout, CsvFile, ExcelFile, _BaseFile
from tests.utils import mock_no_module


@pytest.fixture
def wb_sheets(tmpdir):
    wb = Workbook()
    ws0 = wb.active
    ws0.append([''])
    ws1 = wb.create_sheet('wrong')
    ws1.append(['x'])
    ws1.append(['aa'])
    ws2 = wb.create_sheet('right')
    ws2.append(['x'])
    ws2.append(['a'])
    p = Path(tmpdir, f"{uuid4().hex}.xlsx")
    wb.save(p)
    yield p


def test_exception_message_structure(tmpdir):
    """
    Exception message structure

    This is a very hokey test to check the behavior of the exception message
    structure. This should be improved. This is going to break all the time as
    rule messages and various minor details get changed in the underlying
    classes.
    """
    hid = uuid4().hex[:5]
    p = Path(tmpdir, hid)
    p.write_text('\n'.join(["c1,c2", "1,23", "1,1"]))
    msg = [
        f" - File: {hid}",
        "   - Row: 2",
        "     - Cell: 2,2 (c2)",
        "       - MaxDigit: must have no more than 1 digit characters",
        "   - Column: 1 (c1)",
        "     - Unique: values must be unique",
    ]
    msg = '\n'.join(msg)
    lay = Layout({'c1': Integer(1, rules=[Unique()]), 'c2': Integer(1)})
    try:
        CsvFile(lay).check(p)
    except AssertionError as ae:
        print(ae)
        print(msg)
        assert str(ae).endswith(msg)


def test_documentation():
    """ An equally silly test of documentation output """
    layout = Layout({'x': Field()})
    expected = ' - **x**\n   - cannot be empty/blank'
    assert layout.documentation() == expected


@pytest.mark.parametrize('choice, valid', [
    ('md', True),
    ('html', True),
    ('yyz', False)
])
def test_documentation_types(choice, valid):
    layout = Layout({'x': Field()})
    if valid:
        assert isinstance(layout.documentation(doc_type=choice), str)
    else:
        with pytest.raises(TypeError):
            layout.documentation(doc_type=choice)


@pytest.mark.parametrize('choice, valid', [
    ('md', True),
    ('html', True),
    ('yyz', False)
])
@pytest.mark.parametrize('valid_file', [False, True])
def test_file_output_types(choice, valid, valid_file, tmpdir):
    layout = Layout({'x': Integer(1)})
    p = Path(tmpdir, 'test_file_output_types.csv')
    if valid_file:
        p.write_text('x\n1\n')
    else:
        p.write_text('x\nx\n')

    if valid:
        assert isinstance(CsvFile(layout).check(p, choice), str)
    else:
        with pytest.raises(TypeError):
            CsvFile(layout).check(p, choice)


def test_no_excel(mocker):
    mocker.patch('builtins.__import__', wraps=__import__, side_effect=mock_no_module('openpyxl'))
    with pytest.raises(ModuleNotFoundError):
        ExcelFile(Layout({'x': Integer(1)}))


def test_base_file_stubs():
    assert not _BaseFile(Layout({'x': Integer(1)}))._rows(Path('x'))


def test_excel_wrong_sheet(wb_sheets):
    lay = Layout({'x': Text(1)})
    with pytest.raises(AssertionError):
        ExcelFile(lay, sheet='wrong').check(wb_sheets)


def test_excel_right_sheet(wb_sheets):
    lay = Layout({'x': Text(1)})
    assert not ExcelFile(lay, sheet='right').check(wb_sheets)


def test_file_name_match(tmpdir):
    mock_file = Path(tmpdir, '12345_test_file_report.csv')
    pattern = r'\d{5}_\D*_\D*_report.csv'
    layout = Layout({'x': Integer(1)})
    mock_file.write_text('x\n1\n')
    assert not CsvFile(layout, file_name_pattern=pattern).check(mock_file)



import pytest

from rumydata._base import _BaseRule
from rumydata.exception import UrNotMyDataError


def recurse_subclasses(class_to_recurse):
    def generator(x):
        for y in x.__subclasses__():
            for z in generator(y):
                yield z
        yield x

    return list(generator(class_to_recurse))


@pytest.mark.parametrize('rule', recurse_subclasses(_BaseRule))
def test_rule_exception(rule):
    """ All rules have a UrNotMyDataError type """
    assert issubclass(rule.rule_exception(), UrNotMyDataError)


@pytest.mark.parametrize('rule', recurse_subclasses(_BaseRule))
def test_rule_exception_message(rule):
    """ All rule exceptions are returned as UrNotMyData subclass """
    exc = rule(*rule._default_args)._exception_msg()
    assert issubclass(type(exc), UrNotMyDataError)


@pytest.mark.parametrize('rule', recurse_subclasses(_BaseRule))
def test_rule_default_args(rule):
    """ All rule default arguments are a tuple """
    assert isinstance(rule._default_args, tuple)


@pytest.mark.parametrize('rule', recurse_subclasses(_BaseRule))
def test_rule_explain(rule):
    """ All rule explanations return a string  """
    assert isinstance(rule(*rule._default_args)._explain(), str)


def test_base_prepare():
    r = _BaseRule()._prepare('x')
    assert isinstance(r, tuple)
    assert r[0] == 'x'



"""
rumydata row rules

These rules are applied to entire rows and are generally not intended to be
used directly. They accomplish things like ensuring that the entire row contains
the expected number of values, before attempting to validate individual cells.
"""

from typing import List

from rumydata._base import _BaseRule


class Rule(_BaseRule):
    """ Row Rule """

    def _prepare(self, data: List[str]) -> tuple:
        return data,


class RowLengthLTE(Rule):
    """ Row length less than or equal to Rule """

    _default_args = (1,)

    def __init__(self, columns_length):
        super().__init__()
        self.columns_length = columns_length

    def _evaluator(self):
        return lambda x: len(x) <= self.columns_length

    def _explain(self) -> str:
        return f'row length must be equal to {str(self.columns_length)}, not greater'


class RowLengthGTE(Rule):
    """ Row greater than or equal to Rule """

    _default_args = (1,)

    def __init__(self, columns_length):
        super().__init__()
        self.columns_length = columns_length

    def _evaluator(self):
        return lambda x: len(x) >= self.columns_length

    def _explain(self) -> str:
        return f'row length must be equal to {str(self.columns_length)}, not less'



"""
rumydata base object submodule

This submodule contains the base objects that are used by other modules in this
package. This is not intended for use by end-users.
"""

from typing import List, Union

import rumydata

class _BaseRule:
    """
    Base class for defining data type rules.

    This class contains the default methods that can be used to stub out all of
    the rule types contained in the rules submodule.
    """
    _default_args = tuple()  # a default set of positional args for testing

    def __init__(self):
        pass

    @staticmethod
    def _pre_process(data: tuple) -> tuple:
        """
        Handle data object for rule-type processing prior to evaluation

        This processing is meant to be invoked once per check call, rather than
        once for each rule checked. It handles pre-processing that will apply to
        all rules.

        :param data: an undefined data object
        :return: a tuple, which contains some version of the provided data after
          processing
        """
        return data

    @classmethod
    def rule_exception(cls):
        return type(f'{cls.__name__}Error', (rumydata.exception.UrNotMyDataError,), {})

    def _prepare(self, data) -> tuple:
        """
        Handle data object for rule-specific processing prior to evaluation

        This method solves two problems that exist in the package.

        The first, is that different rule types expect different kinds of data,
        but the subject (e.g. cell, row, column) cannot anticipate what data
        will be needed by each rule. By including a prepare method, and calling
        it to process the data into the evaluator, we can allow multiple types
        of rules which expect different types of data objects to co-exist.

        The second, is that we can perform undifferentiated heavy lifting that
        might be cumbersome in the evaluator method, instead of repeating
        fragments of code multiple times in the evaluator.

        :param data: an undefined data object
        :return: a tuple, which contains some version of the provided data after
          processing
        """
        return data,

    def _evaluator(self):
        """
        Generate a function that will perform evaluation of prepared data.

        The evaluator method provides the caller with a function that can be
        used to evaluate a data object which was pre-processed by this classes
        prepare method. By modifying the lambda function that is returned by
        this method, the user is modifying the actual rule that will be applied
        to any data which is covered by this rule.

        Because the function return by this method is a class method, the user
        has the opportunity to drive change in the returned function by changing
        the arguments of this classes constructor.

        :return: a function which expects to evaluate to True if the value
        provided to the function meets the rule.
        """
        return lambda x: False  # default to failing evaluation if not overwritten

    def _exception_msg(self) -> rumydata.exception.UrNotMyDataError:
        """
        Validation exception message

        Generates an exception from the rule_exception method with a
        sanitized error message that explicitly avoids showing the specific data
        that failed the validation.
        """
        return self.rule_exception()(self._explain())

    def _explain(self) -> str:
        """
        Rule explanation message

        An explanation of the rule that is being applied to the provided data.
        The explanation should be written in a way which clearly links it to the
        evaluation rule. It is very important that the description of the rule
        be clear enough that the user would understand why this particular rule
        failed once they compare the description to their data.
        """
        return "default rule explanation"


class _BaseSubject:
    """
    Base class for defining data subjects.

    This class serves as the basis for the definition of the Field, Layout, and
    File classes. It provides the basic framework for performing checks of rules
    and reporting errors in a way that can be easily collected.
    """

    _default_args = tuple()  # a default set of positional args for testing

    def __init__(self, rules: List[_BaseRule] = None):
        """
        Base subject constructor

        :param rules: a list of rules which should be applied when performing a
        check of data against this object. Rules must all be subclasses of the
        BaseRule.
        """
        self.rules = rules or []
        self.descriptors = {}

    def _check(self, data, rule_type, **kwargs) -> Union[rumydata.exception.UrNotMyDataError, List[rumydata.exception.UrNotMyDataError], None]:
        """
        Check data against specified rule types

        This is the core method used to evaluate data against a subject defined
        by this class.

        :param data: an object which conforms to the `prepare` method of the
            class specified in the rule_type parameter
        :param rule_type: a Rule class belonging to one of the submodules in the
            rules module (e.g. rumydata.rules.cell.Rule). This controls the
            types of rules that the provided data will be checked against.
        :return: a list of any errors that were raised while checking the data.
        """
        errors = []
        if rule_type:
            try:
                data = rule_type._pre_process(data, **kwargs)
            except Exception as e:
                msg = f'raised {e.__class__.__name__} while preprocessing data'
                if rumydata.exception.debug():
                    msg += f' [DEBUG]: {str(e)}'
                return [rumydata.exception.PreProcessingError(msg)]

        for r in self.rules:
            # noinspection PyBroadException
            try:
                if issubclass(type(r), rule_type):
                    x = r._prepare(data)
                    e = r._evaluator()(*x)
                    if not e:
                        errors.append(r._exception_msg())
            except Exception as e:  # get type, and rewrite safe message
                msg = f'raised {e.__class__.__name__} while checking if value {r._explain()}'
                if rumydata.exception.debug():
                    msg += f' [DEBUG]: {str(e)}'
                errors.append(r.rule_exception()(msg))
        return errors

    def _list_errors(self, value, **kwargs) -> List[rumydata.exception.UrNotMyDataError]:
        """
        Flatten nested errors into a list

        This method is a convenience wrapper to perform a check of certain data,
        then flatten the nested errors into a simple list

        :param value: value to be checked
        :return: a list of exceptions raised during check of the provided value
        """
        return list(self._flatten_exceptions(self._check(value, **kwargs)))

    def _has_error(self, value, error, **kwargs) -> bool:
        """

        Check flattened errors for a specific exception type

        This method is a convenience wrapper for testing which determines if the
        nested structure of errors returned by a check includes a specified
        error type.

        :param value: value to be checked.
        :param error: a subclass of UrNotMyDataError which you are checking for.
        :return: a boolean indicator of whether the specified error type was
            returned in the nested structure when checking the provided value.
        """
        # errors must be checked by name since they are generated dynamically for each rule and
        return error.__name__ in [type(x).__name__ for x in self._list_errors(value, **kwargs)]

    def _digest(self) -> List[str]:
        """
        Subject descriptor list

        A method to generate a list of descriptors for this subject, based upon
        the explain method of the rules associated with this subject, as well as
        any additional descriptors set in property of this object

        :return: a list of strings which are used to build a comprehensive
            description of the definition of this subject.
        """
        x = [f'{k}: {v}' if v else k for k, v in self.descriptors.items()]
        y = [x._explain() for x in self.rules]
        return x + y

    @classmethod
    def _flatten_exceptions(cls, error):
        """
        Nested exception generator function

        This method generates a set of nested errors that may exist in a
        UrNotMyData error, as a simple list of errors.

        The design of error reporting creates a nested structure of errors;
        FileError has RowErrors and ColumnErrors, and RowErrors have CellErrors.
        While this is ideal for rolling up errors into a single, comprehensive
        report, this structure makes searching for the presence of particular
        errors difficult.

        This method provides a way to easily search a nested tree of errors for
        a specific error. Essentially, iterates through a list of errors,
        checking to see if any of those errors contain additional errors, then
        continuing to recurse until all errors have been yielded.
        """
        yield error
        for el in error._errors:
            for x in cls._flatten_exceptions(el):
                yield x
